#------------------------------------------------------
namespace eval Object {
    variable Classes {}
    variable ClassInfo

    variable ObjectCount 0
    variable ObjectData
    
    namespace export class
    namespace export method
    namespace export attribute
    namespace export constructor
    namespace export destructor
    namespace export invariant
    namespace export attribute
    namespace export binding
    
    namespace export new
    namespace export delete
    #--------------------------------------------------
    #creates new class
    proc addClass {className {parent ""}} {
        variable Classes
        variable ClassInfo
        
        #check, is class with same name already exsist in the base?
        if {[lsearch $Classes $className] != -1} {
            error "Class $className already exsist"
        }
        #add class to classes list
        lappend Classes $className
        set ClassInfo($className,methods) {}
        set ClassInfo($className,attributes) {}
        set ClassInfo($className,bindings) {}
        
        if {$parent != ""} {
            #create methods table
            set ClassInfo($className,methods) [getMethods $parent]
            
            foreach method $ClassInfo($className,methods) {
                set ClassInfo($className,methods,$method) [getMethodProc $parent $method]
                set ClassInfo($className,methods,$method,precondition) [getPreconditionProc $parent $method]
                set ClassInfo($className,methods,$method,postcondition) [getPostconditionProc $parent $method]
            }
        
            #create attributes table
            set ClassInfo($className,attributes) [getAttributes $parent]
            foreach attrName $ClassInfo($className,attributes) {
                set ClassInfo($className,attributes,$attrName,defValue) [getAttrDefValue $parent $attrName]
            }
            
            #create bindings table
            set ClassInfo($className,bindings) [getBindings $parent]
        } else {
            #all objects has "OnDelete" binding
            lappend ClassInfo($className,bindings) "OnDelete"
        }
        
        set ClassInfo($className,baseClass) $parent
        
        set ClassInfo($className,constructor) ""
        set ClassInfo($className,destructor) ""
        set ClassInfo($className,invariant) ""
    }
    
    #--------------------------------------------------
    # added new method to the class table
    proc addMethod {className method precondition postcondition} {
        variable Classes
        variable ClassInfo

        #check, is class exsist in the base?
        if {[lsearch $Classes $className] == -1} {
            error "Don't know class $className"
        }
        
        set methodProc $className
        append methodProc ::
        append methodProc $method
        set ClassInfo($className,methods,$method) $methodProc
        
        
        if {![info exist ClassInfo($className,methods,$method,precondition)]} {
            set ClassInfo($className,methods,$method,precondition) ""
        }
       
        #set precondition
        if {$precondition != ""} {
            set name $className
            append name ::
            append name $precondition
            set ClassInfo($className,methods,$method,precondition) $name
        }

        #set postcondition
        if {![info exist ClassInfo($className,methods,$method,postcondition)]} {
            set ClassInfo($className,methods,$method,postcondition) ""
        }
       
        if {$postcondition != ""} {
            set name $className
            append name ::
            append name $postcondition
            set ClassInfo($className,methods,$method,postcondition) $name
        }
        
        lappend ClassInfo($className,methods) $method
    }

    #--------------------------------------------------
    # added new attribute
    proc addAttribute {className attrName {defaultValue ""}} {
        variable Classes
        variable ClassInfo
        
        #check, is class exsist in the base?
        if {[lsearch $Classes $className] == -1} {
            error "Don't know class $className"
        }

        #check, is this attribute exsist in the base?
        if {[lsearch ClassInfo($className,attributes) $attrName] != -1} {
            error "Attribute $attrName already exist in $className"
        }
        
        lappend ClassInfo($className,attributes) $attrName
        
        set ClassInfo($className,attributes,$attrName,defValue) $defaultValue
    }
    
    #--------------------------------------------------
    # returns procedure name for the method
    proc getMethodProc {className method} {
        variable Classes
        variable ClassInfo

        return $ClassInfo($className,methods,$method)
    }

    #--------------------------------------------------
    # returns procedure name for the precondition
    proc getPreconditionProc {className method} {
        variable Classes
        variable ClassInfo

        return $ClassInfo($className,methods,$method,precondition)
    }

    #--------------------------------------------------
    # returns procedure name for the postcondition
    proc getPostconditionProc {className method} {
        variable Classes
        variable ClassInfo

        return $ClassInfo($className,methods,$method,postcondition)
    }
    
    #--------------------------------------------------
    #returns list of all methods, declared in the class
    proc getMethods {className} {
        variable Classes
        variable ClassInfo

        #check, is class exsist in the base?
        if {[lsearch $Classes $className] == -1} {
            error "Don't know class $className"
        }
        
        return $ClassInfo($className,methods)
    }

    #--------------------------------------------------
    #returns list of all attributes, declared in the class
    proc getAttributes {className} {
        variable Classes
        variable ClassInfo

        #check, is class exsist in the base?
        if {[lsearch $Classes $className] == -1} {
            error "Don't know class $className"
        }
        
        return $ClassInfo($className,attributes)
    }

    #--------------------------------------------------
    # returns default value for attribute
    proc getAttrDefValue {className attrName} {
        variable ClassInfo

        return $ClassInfo($className,attributes,$attrName,defValue)
    }

    #--------------------------------------------------
    #returns list of all bindings, declared in the class
    proc getBindings {className} {
        variable Classes
        variable ClassInfo

        #check, is class exsist in the base?
        if {[lsearch $Classes $className] == -1} {
            error "Don't know class $className"
        }
        
        return $ClassInfo($className,bindings)
    }
    
    #--------------------------------------------------
    #returns list of classes
    proc getClasses {} {
        variable Classes
        return $Classes
    }
    
    #--------------------------------------------------
    #returns list of inheritance
    proc getInheritList {className} {
        variable ClassInfo
        
        set result {}
        if  {$ClassInfo($className,baseClass) != ""} {
            set result [getInheritList $ClassInfo($className,baseClass)]
        }
        
        lappend result $className
        
        return $result
    }
    
    #--------------------------------------------------
    #base constructor for all object
    proc defaultConstructor {this type} {
        variable ObjectData
        variable ClassInfo
        
        set ObjectData($this,type) $type
        
        #initialize all attributes by default values
        foreach attrName $ClassInfo($type,attributes) {
            set ObjectData($this,attributes,$attrName) $ClassInfo($type,attributes,$attrName,defValue)
        }
        
        #intialize all bindings
        foreach bindName $ClassInfo($type,bindings) {
            set ObjectData($this,bindings,$bindName) {}
        }
        
        set ObjectData($this,deleteActions) ""
    }
    
    #--------------------------------------------------
    #base destructor for all object
    proc defaultDestructor {this} {
        variable ObjectData
        
        set onDeleteBinds $ObjectData($this,bindings,OnDelete)
        
        #perfowrm all delete actions
        foreach action $ObjectData($this,deleteActions) {
            catch {eval $action}
        }
        
        #clear all data
        foreach name [array names ObjectData $this] {
            unset ObjectData($name)
        }
        
        runBindsForList $this "OnDelete" $onDeleteBinds
    }

    #--------------------------------------------------
    #gets object type
    proc getType {this} {
        variable ObjectData
        
        return $ObjectData($this,type)
    }

    #--------------------------------------------------
    #checks, does this object based on specified type
    proc isType {this type} {
        variable ObjectData
        variable ClassInfo
        
        set class $ObjectData($this,type)
        
        while {$class != ""} {
            if {$class == $type} {
                return 1
            }
            
            set class $ClassInfo($class,baseClass)
        }
        
        return 0
    }
    
    #--------------------------------------------------
    #gets attribute value
    proc getAttribute {this attrName} {
        variable ObjectData
        
        return $ObjectData($this,attributes,$attrName)
    }
    
    #--------------------------------------------------
    #sets attribute value
    proc setAttribute {this attrName newValue} {
        variable ObjectData
        
        set oldValue $ObjectData($this,attributes,$attrName)
        
        set ObjectData($this,attributes,$attrName) $newValue
        
        if {$oldValue != $newValue} {
            #run binds
            runBinds $this OnChange$attrName $oldValue $newValue
        }
    }
    
    #--------------------------------------------------
    #waits attribute changing
    proc vwait {this attrName} {
        ::vwait ::Object::ObjectData($this,attributes,$attrName)
    }
    
    #--------------------------------------------------
    #sets constructor
    proc setConstructor {className} {
        variable ClassInfo
        
        set name $className
        append name ::constructor
        set ClassInfo($className,constructor) $name
    }
    
    #--------------------------------------------------
    #gets constructor
    proc getConstructor {className} {
        variable ClassInfo
        
        return $ClassInfo($className,constructor)
    }

    #--------------------------------------------------
    #sets destructor
    proc setDestructor {className} {
        variable ClassInfo
        
        set name $className
        append name ::destructor
        set ClassInfo($className,destructor) $name
    }
    
    #--------------------------------------------------
    #gets destructor
    proc getDestructor {className} {
        variable ClassInfo
        
        return $ClassInfo($className,destructor)
    }
    
    #--------------------------------------------------
    #sets invariant
    proc setInvariant {className} {
        variable ClassInfo
        
        set name $className
        append name ::invariant
        set ClassInfo($className,invariant) $name
    }
    
    #--------------------------------------------------
    #gets invariant
    proc getInvariant {className} {
        variable ClassInfo
        
        return $ClassInfo($className,invariant)
    }

    #------------------------------------------------------
    # Commands for working with bindings
    #------------------------------------------------------
    proc addBinding {className bindName} {
        variable ClassInfo
        variable Classes
        
        #check, is class exsist in the base?
        if {[lsearch $Classes $className] == -1} {
            error "Don't know class $className"
        }

        #check, is this binding exsist in the base?
        if {[lsearch ClassInfo($className,bindings) $bindName] != -1} {
            error "Binding $attrName already exist in $className"
        }
        
        lappend ClassInfo($className,bindings) $bindName
    }
    
    #------------------------------------------------------
    proc bind {this bindName script} {
        variable ObjectData
        variable ClassInfo
        set className $ObjectData($this,type)
        
        if {[lsearch $ClassInfo($className,bindings) $bindName] == -1} {
            error "Unknown bind for $className - $bindName"
        }

        lappend ObjectData($this,bindings,$bindName) $script
        
        return
    }
    
    #------------------------------------------------------
    proc unbind {this bindName script} {
        variable ObjectData
        variable ClassInfo
        set className $ObjectData($this,type)
        
        if {[lsearch $ClassInfo($className,bindings) $bindName] == -1} {
            error "Unknown bind for $className - $bindName"
        }
        
        set index [lsearch $ObjectData($this,bindings,$bindName) $script]
        
        if {$index == -1} {
            error "Here is no bind on $bindName"
        }
        
        set ObjectData($this,bindings,$bindName) \
        [lreplace $ObjectData($this,bindings,$bindName) $index $index]
        
        return
    }
    
    #------------------------------------------------------
    proc runBinds {this bindName args} {
        variable ObjectData
        eval "runBindsForList $this $bindName \{$ObjectData($this,bindings,$bindName)\} $args"
        return
    }


    #------------------------------------------------------
    proc runBindsForList {this bindName bindList args} {
        foreach procName $bindList {
            eval "$procName $bindName $this $args"
        }
        return
    }    
    
    #------------------------------------------------------
    proc addDeleteAction {this action} {
        variable ObjectData
        
        lappend  ObjectData($this,deleteActions) $action
    }
    
    #------------------------------------------------------
    proc killDeleteAction {this action} {
        variable ObjectData
        
        set i [lsearch $ObjectData($this,deleteActions) $action]
        
        if {$i != -1} {
            set ObjectData($this,deleteActions) [lreplace ObjectData($this,deleteActions) $i $i]
        }
    }
    
    #------------------------------------------------------
    # Exported commands
    #------------------------------------------------------
    #declares new class
    proc class {className args} {
        switch -- [llength $args] {
            1 {
                set parent ""
                set body [lindex $args 0]
            }
            
            2 {
                set parent [lindex $args 0]
                set body [lindex $args 1]
            }
            
            default {
                error "Incorrect Parametrs. Shoould be \"class className ?parent? body\""
            }
        }
        
        addClass $className $parent
        
        eval "namespace eval $className \{\
            variable className $className \
            $body \
        \}"
    }

    #------------------------------------------------------
    #declares new method
    proc method {methodName arguments body args} {
        set usePrecondition 0
        set usePostcondition 0
        set precondition ""
        set postcondition ""
        
        for {set i 0} {$i < [llength $args]} {incr i} {
            set argName [lindex $args $i]
            incr i
            set argValue [lindex $args $i]
            
            switch -- $argName {
                "-precondition" {
                    set precondition $argValue
                    set usePrecondition 1
                }
                "-postcondition" {
                    set postcondition $argValue
                    set usePostcondition 1
                }
                default {
                    "Unknown argument $argName. Should be -precondition or -postcondition"
                }
            }
        }
        
        eval "uplevel 1 \{
        
        #create precondition
        if \{$usePrecondition != 0\} \{
        set name $methodName
        append name _precondition
        proc \$name \{this $arguments \} \{
            $precondition
        \}
        set precondition \$name
        \} else \{
        set precondition \"\"
        \}
        
        #create postcondition
        if \{$usePostcondition != 0\} \{
        set name $methodName
        append name _postcondition
        proc \$name \{this $arguments \} \{
            $postcondition
        \} 
        set postcondition \$name
        \} else \{
        set postcondition \"\"
        \}
        
        Object::addMethod \$className $methodName \$precondition \$postcondition
        proc $methodName \{ this $arguments \} \{ $body \}
        \}"
        
    }
    
    #------------------------------------------------------
    #declares new attribute
    proc attribute {attrName {defaultValue ""}} {
        eval "uplevel 1 \{
        Object::addAttribute \$className $attrName \"$defaultValue\"
        
        binding OnChange$attrName
        
        method get$attrName \{\} \{
            return \[Object::getAttribute \$this $attrName\]
        \}

        method set$attrName \{newValue\} \{
            Object::setAttribute \$this $attrName \$newValue
        \}
        
        \}"
    }

    #------------------------------------------------------
    proc binding {bindName} {
        eval "uplevel 1 \{
        Object::addBinding \$className $bindName
        \}"
    }
    
    #------------------------------------------------------
    #declares constructor
    proc constructor {body} {
        eval "uplevel 1 \{
        Object::setConstructor \$className
        proc constructor \{ this args\} \{ $body \}
        \}"
    }

    #------------------------------------------------------
    #declares destructor
    proc destructor {body} {
        eval "uplevel 1 \{
        Object::setDestructor \$className
        proc destructor \{ this \} \{ $body \}
        \}"
    }

    #------------------------------------------------------
    #invariant destructor
    proc invariant {body} {
        eval "uplevel 1 \{
        Object::setInvariant \$className
        proc invariant \{ this \} \{ $body \}
        \}"
    }
    
    #------------------------------------------------------
    #creates new object
    proc new {className args} {
        global ObjectDebug
        
        variable ObjectCount
        
        set objName obj$ObjectCount
        incr ObjectCount
        
        if {[info exist ObjectDebug] && ($ObjectDebug == "1")} {
            #create procedure (debug variant)
            proc ::$objName {method args} "\
            set invariant \[::Object::getInvariant $className\]; \
            if \{\$invariant != \"\"\} \{ \
                ::Object::\$invariant $objName \
            \}; \
            set precondition \[::Object::getPreconditionProc $className \$method\]; \
            if \{\$precondition != \"\"\} \{ \
                eval ::Object::\$precondition $objName \$args \
            \}; \
            set methodProc \[Object::getMethodProc $className \$method \];\
            set result \[eval ::Object::\$methodProc $objName \$args \] ;\
            set postcondition \[::Object::getPostconditionProc $className \$method\]; \
            if \{\$postcondition != \"\"\} \{ \
                eval ::Object::\$postcondition $objName \$args \
            \}; \
            return \$result; \
            "
        } else {
            #create procedure (nondebug variant)
            proc ::$objName {method args} "\
            set methodProc \[Object::getMethodProc $className \$method \];\
            return \[eval ::Object::\$methodProc $objName \$args \] ;\
            "
        }
        
        #call constructors for object
        defaultConstructor $objName $className
        set classes [getInheritList $className]
        foreach class $classes {
            set constructor [getConstructor $class]
            
            if {$constructor != ""} {
                #call constructor
                if {$class != $className} {
                    $constructor $objName
                } else {
                    eval $constructor $objName $args
                }
            }
        }
        
        return $objName
    }

    #------------------------------------------------------
    #deletes object
    proc delete {obj} {
        set className [getType $obj]
        #calls destructors
        set classes [getInheritList $className]
        
        for {set i [expr {[llength $classes]-1}]} {$i >= 0} { incr i -1} {
            set class [lindex $classes $i]
            set destructor [getDestructor $class]
            
            if {$destructor != ""} {
                $destructor $obj
            }
        }
        
        defaultDestructor $obj
        
        #delete proc
        rename $obj ""
    }
}
